@@ -51,7 +51,7 @@
 #include "misc.h"
 #include "gdmwm.h"
 
-static const gchar RCSid[]="$Id: gdmchooser.c,v 1.28 2001/08/04 23:28:52 jirka Exp $";
+static const gchar RCSid[]="$Id: gdmchooser.c,v 1.29 2001/08/25 21:46:15 jirka Exp $";
 
 /* set the DOING_GDM_DEVELOPMENT env variable if you want to
  * search for the glade file in the current dir and not the system
@@ -910,7 +910,7 @@ gdm_chooser_choose_host (const char *hostname)
       if ((fd = socket (PF_INET, SOCK_STREAM, 0)) == -1)
 	      gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't create response socket.");
 
-      if (connect (fd, (struct sockaddr_in *) &in_addr, sizeof (in_addr)) == -1)
+      if (connect (fd, (struct sockaddr *) &in_addr, sizeof (in_addr)) == -1)
 	      gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't connect to xdm.");
       buffer.data = (BYTE *) buf;
       buffer.size = sizeof (buf);
